"""
This file is used by test_apps.py to make sure we can exclude imported
``Table`` subclasses when using ``table_finder``.
"""

from piccolo.apps.user.tables import BaseUser
from piccolo.columns.column_types import ForeignKey, Varchar
from piccolo.table import Table


class Musician(Table):
    name = Varchar()
    user = ForeignKey(BaseUser)



from unittest import TestCase

from piccolo.columns import Integer, Varchar
from piccolo.table import Table


class Band(Table):
    name = Varchar()
    popularity = Integer()


class TestCount(TestCase):
    def setUp(self) -> None:
        Band.create_table().run_sync()

    def tearDown(self) -> None:
        Band.alter().drop_table().run_sync()

    def test_count(self):
        Band.insert(
            Band(name="Pythonistas", popularity=10),
            Band(name="Rustaceans", popularity=10),
            Band(name="C-Sharps", popularity=5),
        ).run_sync()

        response = Band.count().run_sync()

        self.assertEqual(response, 3)

    def test_count_where(self):
        Band.insert(
            Band(name="Pythonistas", popularity=10),
            Band(name="Rustaceans", popularity=10),
            Band(name="C-Sharps", popularity=5),
        ).run_sync()

        response = Band.count().where(Band.popularity == 10).run_sync()

        self.assertEqual(response, 2)

    def test_count_distinct(self):
        Band.insert(
            Band(name="Pythonistas", popularity=10),
            Band(name="Rustaceans", popularity=10),
            Band(name="C-Sharps", popularity=5),
            Band(name="Fortranists", popularity=2),
        ).run_sync()

        response = Band.count(distinct=[Band.popularity]).run_sync()

        self.assertEqual(response, 3)

        # Test the method also works
        response = Band.count().distinct([Band.popularity]).run_sync()
        self.assertEqual(response, 3)

    def test_count_distinct_multiple(self):
        Band.insert(
            Band(name="Pythonistas", popularity=10),
            Band(name="Pythonistas", popularity=10),
            Band(name="Rustaceans", popularity=10),
            Band(name="C-Sharps", popularity=5),
            Band(name="Fortranists", popularity=2),
        ).run_sync()

        response = Band.count(distinct=[Band.name, Band.popularity]).run_sync()

        self.assertEqual(response, 4)

    def test_value_error(self):
        """
        Make sure specifying `column` and `distinct` raises an error.
        """
        with self.assertRaises(ValueError):
            Band.count(
                column=Band.name, distinct=[Band.name, Band.popularity]
            ).run_sync()



from __future__ import annotations

import contextvars
import typing as t
from dataclasses import dataclass

from typing_extensions import Self

from piccolo.engine.base import (
    BaseAtomic,
    BaseBatch,
    BaseTransaction,
    Engine,
    validate_savepoint_name,
)
from piccolo.engine.exceptions import TransactionError
from piccolo.query.base import DDL, Query
from piccolo.querystring import QueryString
from piccolo.utils.lazy_loader import LazyLoader
from piccolo.utils.sync import run_sync
from piccolo.utils.warnings import Level, colored_warning

asyncpg = LazyLoader("asyncpg", globals(), "asyncpg")

if t.TYPE_CHECKING:  # pragma: no cover
    from asyncpg.connection import Connection
    from asyncpg.cursor import Cursor
    from asyncpg.pool import Pool
    from asyncpg.transaction import Transaction


@dataclass
class AsyncBatch(BaseBatch):
    connection: Connection
    query: Query
    batch_size: int

    # Set internally
    _transaction: t.Optional[Transaction] = None
    _cursor: t.Optional[Cursor] = None

    @property
    def cursor(self) -> Cursor:
        if not self._cursor:
            raise ValueError("_cursor not set")
        return self._cursor

    @property
    def transaction(self) -> Transaction:
        if not self._transaction:
            raise ValueError("The transaction can't be found.")
        return self._transaction

    async def next(self) -> t.List[t.Dict]:
        data = await self.cursor.fetch(self.batch_size)
        return await self.query._process_results(data)

    def __aiter__(self: Self) -> Self:
        return self

    async def __anext__(self) -> t.List[t.Dict]:
        response = await self.next()
        if response == []:
            raise StopAsyncIteration()
        return response

    async def __aenter__(self: Self) -> Self:
        self._transaction = self.connection.transaction()
        await self._transaction.start()
        querystring = self.query.querystrings[0]
        template, template_args = querystring.compile_string()

        self._cursor = await self.connection.cursor(template, *template_args)
        return self

    async def __aexit__(self, exception_type, exception, traceback):
        if exception:
            await self.transaction.rollback()
        else:
            await self.transaction.commit()

        await self.connection.close()

        return exception is not None


###############################################################################


class Atomic(BaseAtomic):
    """
    This is useful if you want to build up a transaction programatically, by
    adding queries to it.

    Usage::

        transaction = engine.atomic()
        transaction.add(Foo.create_table())

        # Either:
        transaction.run_sync()
        await transaction.run()

    """

    __slots__ = ("engine", "queries")

    def __init__(self, engine: PostgresEngine):
        self.engine = engine
        self.queries: t.List[t.Union[Query, DDL]] = []

    def add(self, *query: t.Union[Query, DDL]):
        self.queries += list(query)

    async def run(self):
        from piccolo.query.methods.objects import Create, GetOrCreate

        try:
            async with self.engine.transaction():
                for query in self.queries:
                    if isinstance(query, (Query, DDL, Create, GetOrCreate)):
                        await query.run()
                    else:
                        raise ValueError("Unrecognised query")
            self.queries = []
        except Exception as exception:
            self.queries = []
            raise exception from exception

    def run_sync(self):
        return run_sync(self.run())

    def __await__(self):
        return self.run().__await__()


###############################################################################


class Savepoint:
    def __init__(self, name: str, transaction: PostgresTransaction):
        self.name = name
        self.transaction = transaction

    async def rollback_to(self):
        validate_savepoint_name(self.name)
        await self.transaction.connection.execute(
            f"ROLLBACK TO SAVEPOINT {self.name}"
        )

    async def release(self):
        validate_savepoint_name(self.name)
        await self.transaction.connection.execute(
            f"RELEASE SAVEPOINT {self.name}"
        )


class PostgresTransaction(BaseTransaction):
    """
    Used for wrapping queries in a transaction, using a context manager.
    Currently it's async only.

    Usage::

        async with engine.transaction():
            # Run some queries:
            await Band.select().run()

    """

    __slots__ = (
        "engine",
        "transaction",
        "context",
        "connection",
        "_savepoint_id",
        "_parent",
        "_committed",
        "_rolled_back",
    )

    def __init__(self, engine: PostgresEngine, allow_nested: bool = True):
        """
        :param allow_nested:
            If ``True`` then if we try creating a new transaction when another
            is already active, we treat this as a no-op::

                async with DB.transaction():
                    async with DB.transaction():
                        pass

            If we want to disallow this behaviour, then setting
            ``allow_nested=False`` will cause a ``TransactionError`` to be
            raised.

        """
        self.engine = engine
        current_transaction = self.engine.current_transaction.get()

        self._savepoint_id = 0
        self._parent = None
        self._committed = False
        self._rolled_back = False

        if current_transaction:
            if allow_nested:
                self._parent = current_transaction
            else:
                raise TransactionError(
                    "A transaction is already active - nested transactions "
                    "aren't allowed."
                )

    async def __aenter__(self) -> PostgresTransaction:
        if self._parent is not None:
            return self._parent

        self.connection = await self.get_connection()
        self.transaction = self.connection.transaction()
        await self.begin()
        self.context = self.engine.current_transaction.set(self)
        return self

    async def get_connection(self):
        if self.engine.pool:
            return await self.engine.pool.acquire()
        else:
            return await self.engine.get_new_connection()

    async def begin(self):
        await self.transaction.start()

    async def commit(self):
        await self.transaction.commit()
        self._committed = True

    async def rollback(self):
        await self.transaction.rollback()
        self._rolled_back = True

    async def rollback_to(self, savepoint_name: str):
        """
        Used to rollback to a savepoint just using the name.
        """
        await Savepoint(name=savepoint_name, transaction=self).rollback_to()

    ###########################################################################

    def get_savepoint_id(self) -> int:
        self._savepoint_id += 1
        return self._savepoint_id

    async def savepoint(self, name: t.Optional[str] = None) -> Savepoint:
        name = name or f"savepoint_{self.get_savepoint_id()}"
        validate_savepoint_name(name)
        await self.connection.execute(f"SAVEPOINT {name}")
        return Savepoint(name=name, transaction=self)

    ###########################################################################

    async def __aexit__(self, exception_type, exception, traceback) -> bool:
        if self._parent:
            return exception is None

        if exception:
            # The user may have manually rolled it back.
            if not self._rolled_back:
                await self.rollback()
        else:
            # The user may have manually committed it.
            if not self._committed and not self._rolled_back:
                await self.commit()

        if self.engine.pool:
            await self.engine.pool.release(self.connection)
        else:
            await self.connection.close()

        self.engine.current_transaction.reset(self.context)

        return exception is None


###############################################################################


class PostgresEngine(Engine[PostgresTransaction]):
    """
    Used to connect to PostgreSQL.

    :param config:
        The config dictionary is passed to the underlying database adapter,
        asyncpg. Common arguments you're likely to need are:

        * host
        * port
        * user
        * password
        * database

        For example, ``{'host': 'localhost', 'port': 5432}``.

        See the `asyncpg docs <https://magicstack.github.io/asyncpg/current/api/index.html#connection>`_
        for all available options.

    :param extensions:
        When the engine starts, it will try and create these extensions
        in Postgres. If you're using a read only database, set this value to an
        empty tuple ``()``.

    :param log_queries:
        If ``True``, all SQL and DDL statements are printed out before being
        run. Useful for debugging.

    :param log_responses:
        If ``True``, the raw response from each query is printed out. Useful
        for debugging.

    :param extra_nodes:
        If you have additional database nodes (e.g. read replicas) for the
        server, you can specify them here. It's a mapping of a memorable name
        to a ``PostgresEngine`` instance. For example::

            DB = PostgresEngine(
                config={'database': 'main_db'},
                extra_nodes={
                    'read_replica_1': PostgresEngine(
                        config={
                            'database': 'main_db',
                            host: 'read_replicate.my_db.com'
                        },
                        extensions=()
                    )
                }
            )

        Note how we set ``extensions=()``, because it's a read only database.

        When executing a query, you can specify one of these nodes instead
        of the main database. For example::

            >>> await MyTable.select().run(node="read_replica_1")

    """  # noqa: E501

    __slots__ = (
        "config",
        "extensions",
        "extra_nodes",
        "pool",
    )

    def __init__(
        self,
        config: t.Dict[str, t.Any],
        extensions: t.Sequence[str] = ("uuid-ossp",),
        log_queries: bool = False,
        log_responses: bool = False,
        extra_nodes: t.Optional[t.Mapping[str, PostgresEngine]] = None,
    ) -> None:
        if extra_nodes is None:
            extra_nodes = {}

        self.config = config
        self.extensions = extensions
        self.log_queries = log_queries
        self.log_responses = log_responses
        self.extra_nodes = extra_nodes
        self.pool: t.Optional[Pool] = None
        database_name = config.get("database", "Unknown")
        self.current_transaction = contextvars.ContextVar(
            f"pg_current_transaction_{database_name}", default=None
        )
        super().__init__(
            engine_type="postgres",
            log_queries=log_queries,
            log_responses=log_responses,
            min_version_number=10,
        )

    @staticmethod
    def _parse_raw_version_string(version_string: str) -> float:
        """
        The format of the version string isn't always consistent. Sometimes
        it's just the version number e.g. '9.6.18', and sometimes
        it contains specific build information e.g.
        '12.4 (Ubuntu 12.4-0ubuntu0.20.04.1)'. Just extract the major and
        minor version numbers.
        """
        version_segment = version_string.split(" ")[0]
        major, minor = version_segment.split(".")[:2]
        return float(f"{major}.{minor}")

    async def get_version(self) -> float:
        """
        Returns the version of Postgres being run.
        """
        try:
            response: t.Sequence[t.Dict] = await self._run_in_new_connection(
                "SHOW server_version"
            )
        except ConnectionRefusedError as exception:
            # Suppressing the exception, otherwise importing piccolo_conf.py
            # containing an engine will raise an ImportError.
            colored_warning(f"Unable to connect to database - {exception}")
            return 0.0
        else:
            version_string = response[0]["server_version"]
            return self._parse_raw_version_string(
                version_string=version_string
            )

    def get_version_sync(self) -> float:
        return run_sync(self.get_version())

    async def prep_database(self):
        for extension in self.extensions:
            try:
                await self._run_in_new_connection(
                    f'CREATE EXTENSION IF NOT EXISTS "{extension}"',
                )
            except asyncpg.exceptions.InsufficientPrivilegeError:
                colored_warning(
                    f"=> Unable to create {extension} extension - some "
                    "functionality may not behave as expected. Make sure "
                    "your database user has permission to create "
                    "extensions, or add it manually using "
                    f'`CREATE EXTENSION "{extension}";`',
                    level=Level.medium,
                )

    ###########################################################################
    # These typos existed in the codebase for a while, so leaving these proxy
    # methods for now to ensure backwards compatibility.

    async def start_connnection_pool(self, **kwargs) -> None:
        colored_warning(
            "`start_connnection_pool` is a typo - please change it to "
            "`start_connection_pool`.",
            category=DeprecationWarning,
        )
        return await self.start_connection_pool()

    async def close_connnection_pool(self, **kwargs) -> None:
        colored_warning(
            "`close_connnection_pool` is a typo - please change it to "
            "`close_connection_pool`.",
            category=DeprecationWarning,
        )
        return await self.close_connection_pool()

    ###########################################################################

    async def start_connection_pool(self, **kwargs) -> None:
        if self.pool:
            colored_warning(
                "A pool already exists - close it first if you want to create "
                "a new pool.",
            )
        else:
            config = dict(self.config)
            config.update(**kwargs)
            self.pool = await asyncpg.create_pool(**config)

    async def close_connection_pool(self) -> None:
        if self.pool:
            await self.pool.close()
            self.pool = None
        else:
            colored_warning("No pool is running.")

    ###########################################################################

    async def get_new_connection(self) -> Connection:
        """
        Returns a new connection - doesn't retrieve it from the pool.
        """
        return await asyncpg.connect(**self.config)

    ###########################################################################

    async def batch(
        self,
        query: Query,
        batch_size: int = 100,
        node: t.Optional[str] = None,
    ) -> AsyncBatch:
        """
        :param query:
            The database query to run.
        :param batch_size:
            How many rows to fetch on each iteration.
        :param node:
            Which node to run the query on (see ``extra_nodes``). If not
            specified, it runs on the main Postgres node.
        """
        engine: t.Any = self.extra_nodes.get(node) if node else self
        connection = await engine.get_new_connection()
        return AsyncBatch(
            connection=connection, query=query, batch_size=batch_size
        )

    ###########################################################################

    async def _run_in_pool(
        self, query: str, args: t.Optional[t.Sequence[t.Any]] = None
    ):
        if args is None:
            args = []
        if not self.pool:
            raise ValueError("A pool isn't currently running.")

        async with self.pool.acquire() as connection:
            response = await connection.fetch(query, *args)

        return response

    async def _run_in_new_connection(
        self, query: str, args: t.Optional[t.Sequence[t.Any]] = None
    ):
        if args is None:
            args = []
        connection = await self.get_new_connection()

        try:
            results = await connection.fetch(query, *args)
        except asyncpg.exceptions.PostgresError as exception:
            await connection.close()
            raise exception

        await connection.close()
        return results

    async def run_querystring(
        self, querystring: QueryString, in_pool: bool = True
    ):
        query, query_args = querystring.compile_string(
            engine_type=self.engine_type
        )

        query_id = self.get_query_id()

        if self.log_queries:
            self.print_query(query_id=query_id, query=querystring.__str__())

        # If running inside a transaction:
        current_transaction = self.current_transaction.get()
        if current_transaction:
            response = await current_transaction.connection.fetch(
                query, *query_args
            )
        elif in_pool and self.pool:
            response = await self._run_in_pool(query, query_args)
        else:
            response = await self._run_in_new_connection(query, query_args)

        if self.log_responses:
            self.print_response(query_id=query_id, response=response)

        return response

    async def run_ddl(self, ddl: str, in_pool: bool = True):
        query_id = self.get_query_id()

        if self.log_queries:
            self.print_query(query_id=query_id, query=ddl)

        # If running inside a transaction:
        current_transaction = self.current_transaction.get()
        if current_transaction:
            response = await current_transaction.connection.fetch(ddl)
        elif in_pool and self.pool:
            response = await self._run_in_pool(ddl)
        else:
            response = await self._run_in_new_connection(ddl)

        if self.log_responses:
            self.print_response(query_id=query_id, response=response)

        return response

    def atomic(self) -> Atomic:
        return Atomic(engine=self)

    def transaction(self, allow_nested: bool = True) -> PostgresTransaction:
        return PostgresTransaction(engine=self, allow_nested=allow_nested)



import decimal
from unittest import TestCase

from piccolo.columns import ForeignKey, Varchar
from piccolo.columns.readable import Readable
from piccolo.table import Table, create_db_tables_sync, drop_db_tables_sync
from tests.base import engine_is
from tests.example_apps.music.tables import (
    Band,
    Concert,
    Manager,
    Ticket,
    Venue,
)


class ThingOne(Table):
    name = Varchar(length=300, null=False)


class ThingTwo(Table):
    name = Varchar(length=300, null=False)
    thing_one = ForeignKey(references=ThingOne)


class ThingThree(Table):
    name = Varchar(length=300, null=False)
    thing_two = ForeignKey(references=ThingTwo)

    @classmethod
    def get_readable(cls):
        return Readable(
            template="three name: %s - two name: %s - one name: %s",
            columns=[
                cls.name,
                cls.thing_two.name,
                cls.thing_two._.thing_one._.name,
            ],
        )


class ThingFour(Table):
    name = Varchar(length=300, null=False)
    thing_three = ForeignKey(references=ThingThree)


TABLES = [
    Band,
    Concert,
    Manager,
    Venue,
    Ticket,
    ThingOne,
    ThingTwo,
    ThingThree,
    ThingFour,
]


class TestGetRelatedReadable(TestCase):
    def setUp(self):
        create_db_tables_sync(*TABLES)

        manager_1 = Manager.objects().create(name="Guido").run_sync()
        manager_2 = Manager.objects().create(name="Graydon").run_sync()

        band_1 = (
            Band.objects()
            .create(name="Pythonistas", manager=manager_1)
            .run_sync()
        )
        band_2 = (
            Band.objects()
            .create(name="Rustaceans", manager=manager_2)
            .run_sync()
        )
        venue = (
            Venue.objects()
            .create(name="Royal Albert Hall", capacity=5900)
            .run_sync()
        )
        concert = (
            Concert.objects()
            .create(venue=venue, band_1=band_1, band_2=band_2)
            .run_sync()
        )
        Ticket.objects().create(
            price=decimal.Decimal(50.0), concert=concert
        ).run_sync()

        thing_one = ThingOne.insert(ThingOne(name="thing_one")).run_sync()
        thing_two = ThingTwo.insert(
            ThingTwo(name="thing_two", thing_one=thing_one[0]["id"])
        ).run_sync()
        thing_three = ThingThree.insert(
            ThingThree(name="thing_three", thing_two=thing_two[0]["id"])
        ).run_sync()
        ThingFour.insert(
            ThingFour(name="thing_four", thing_three=thing_three[0]["id"])
        ).run_sync()

    def tearDown(self):
        drop_db_tables_sync(*TABLES)

    def test_get_related_readable(self):
        """
        Make sure you can get the `Readable` representation for related object
        from another object instance.
        """
        response = Band.select(
            Band.name, Band._get_related_readable(Band.manager)
        ).run_sync()

        self.assertEqual(
            response,
            [
                {"name": "Pythonistas", "manager_readable": "Guido"},
                {"manager_readable": "Graydon", "name": "Rustaceans"},
            ],
        )

        # Now try something much more complex.
        response = Ticket.select(
            Ticket.id,
            Ticket._get_related_readable(Ticket.concert),
        ).run_sync()

        if engine_is("cockroach"):
            self.assertEqual(
                response,
                [
                    {
                        "id": response[0]["id"],
                        "concert_readable": (
                            "Pythonistas and Rustaceans at Royal Albert Hall, "
                            "capacity 5900"
                        ),
                    }
                ],
            )
        else:
            self.assertEqual(
                response,
                [
                    {
                        "id": 1,
                        "concert_readable": (
                            "Pythonistas and Rustaceans at Royal Albert Hall, "
                            "capacity 5900"
                        ),
                    }
                ],
            )

        # A really complex references chain from Piccolo Admin issue #170
        response = ThingFour.select(
            ThingFour._get_related_readable(ThingFour.thing_three)
        ).run_sync()
        self.assertEqual(
            response,
            [
                {
                    "thing_three_readable": (
                        "three name: thing_three - two name: thing_two - one name: thing_one"  # noqa: E501
                    )
                }
            ],
        )



import typing as t

ElementType = t.TypeVar("ElementType")


def flatten(
    items: t.Sequence[t.Union[ElementType, t.List[ElementType]]]
) -> t.List[ElementType]:
    """
    Takes a sequence of elements, and flattens it out. For example::

        >>> flatten(['a', ['b', 'c']])
        ['a', 'b', 'c']

    We need this for situations like this::

        await Band.select(Band.name, Band.manager.all_columns())

    """
    _items: t.List[ElementType] = []
    for item in items:
        if isinstance(item, list):
            _items.extend(item)
        else:
            _items.append(item)

    return _items


def batch(
    data: t.List[ElementType], chunk_size: int
) -> t.List[t.List[ElementType]]:
    """
    Breaks the list down into sublists of the given ``chunk_size``. The last
    sublist may have fewer elements than ``chunk_size``::

        >>> batch(['a', 'b', 'c'], 2)
        [['a', 'b'], ['c']]

    :param data:
        The data to break up into sublists.
    :param chunk_size:
        How large each sublist should be.

    """
    # TODO: Replace with itertools.batch when available:
    # https://docs.python.org/3.12/library/itertools.html#itertools.batched

    if chunk_size <= 0:
        raise ValueError("`chunk_size` must be greater than 0.")

    row_count = len(data)

    iterations = int(row_count / chunk_size)
    if row_count % chunk_size > 0:
        iterations += 1

    return [
        data[(i * chunk_size) : ((i + 1) * chunk_size)]  # noqa: E203
        for i in range(0, iterations)
    ]



from piccolo.query.functions.string import Upper
from tests.example_apps.music.tables import Band

from .base import BandTest


class TestUpperFunction(BandTest):

    def test_column(self):
        """
        Make sure we can uppercase a column's value.
        """
        response = Band.select(Upper(Band.name)).run_sync()
        self.assertListEqual(response, [{"upper": "PYTHONISTAS"}])

    def test_alias(self):
        response = Band.select(Upper(Band.name, alias="name")).run_sync()
        self.assertListEqual(response, [{"name": "PYTHONISTAS"}])

    def test_joined_column(self):
        """
        Make sure we can uppercase a column's value from a joined table.
        """
        response = Band.select(Upper(Band.manager._.name)).run_sync()
        self.assertListEqual(response, [{"upper": "GUIDO"}])



import typing as t
from unittest import TestCase

from piccolo.table import Table, create_db_tables_sync, drop_db_tables_sync
from tests.example_apps.music.tables import Band, Manager


class FunctionTest(TestCase):
    tables: t.List[t.Type[Table]]

    def setUp(self) -> None:
        create_db_tables_sync(*self.tables)

    def tearDown(self) -> None:
        drop_db_tables_sync(*self.tables)


class BandTest(FunctionTest):
    tables = [Band, Manager]

    def setUp(self) -> None:
        super().setUp()

        manager = Manager({Manager.name: "Guido"})
        manager.save().run_sync()

        band = Band(
            {
                Band.name: "Pythonistas",
                Band.manager: manager,
                Band.popularity: 1000,
            }
        )
        band.save().run_sync()



import datetime
from unittest import TestCase

from piccolo.columns.column_types import Timestamp
from piccolo.columns.defaults.timestamp import TimestampNow
from piccolo.table import Table


class MyTable(Table):
    created_on = Timestamp()


class MyTableDefault(Table):
    """
    A table containing all of the possible `default` arguments for
    `Timestamp`.
    """

    created_on = Timestamp(default=TimestampNow())


class TestTimestamp(TestCase):
    def setUp(self):
        MyTable.create_table().run_sync()

    def tearDown(self):
        MyTable.alter().drop_table().run_sync()

    def test_timestamp(self):
        """
        Make sure a datetime can be stored and retrieved.
        """
        created_on = datetime.datetime.now()
        row = MyTable(created_on=created_on)
        row.save().run_sync()

        result = MyTable.objects().first().run_sync()
        assert result is not None
        self.assertEqual(result.created_on, created_on)

    def test_timezone_aware(self):
        """
        Raise an error if a timezone aware datetime is given as a default.
        """
        with self.assertRaises(ValueError):
            Timestamp(default=datetime.datetime.now(tz=datetime.timezone.utc))


class TestTimestampDefault(TestCase):
    def setUp(self):
        MyTableDefault.create_table().run_sync()

    def tearDown(self):
        MyTableDefault.alter().drop_table().run_sync()

    def test_timestamp(self):
        """
        Make sure the default values get created correctly.
        """
        created_on = datetime.datetime.now()
        row = MyTableDefault()
        row.save().run_sync()

        result = MyTableDefault.objects().first().run_sync()
        assert result is not None
        self.assertLess(
            result.created_on - created_on, datetime.timedelta(seconds=1)
        )
        self.assertIsNone(result.created_on.tzinfo)



import string
from unittest import TestCase

from piccolo.utils.list import batch, flatten


class TestFlatten(TestCase):
    def test_flatten(self):
        self.assertListEqual(flatten(["a", ["b", "c"]]), ["a", "b", "c"])


class TestBatch(TestCase):
    def test_batch(self):
        self.assertListEqual(
            batch([i for i in string.ascii_lowercase], chunk_size=5),
            [
                ["a", "b", "c", "d", "e"],
                ["f", "g", "h", "i", "j"],
                ["k", "l", "m", "n", "o"],
                ["p", "q", "r", "s", "t"],
                ["u", "v", "w", "x", "y"],
                ["z"],
            ],
        )

    def test_zero(self):
        with self.assertRaises(ValueError):
            batch([1, 2, 3], chunk_size=0)



from __future__ import annotations

import typing as t

from piccolo.apps.migrations.commands.base import BaseMigrationManager
from piccolo.apps.migrations.tables import Migration


class CleanMigrationManager(BaseMigrationManager):
    def __init__(self, app_name: str, auto_agree: bool = False):
        self.app_name = app_name
        self.auto_agree = auto_agree
        super().__init__()

    def get_migration_ids_to_remove(self) -> t.List[str]:
        """
        Returns a list of migration ID strings, which are rows in the table,
        but don't have a corresponding migration module on disk.
        """
        app_config = self.get_app_config(app_name=self.app_name)

        migration_module_dict = self.get_migration_modules(
            folder_path=app_config.resolved_migrations_folder_path
        )

        # The migration IDs which are in migration modules.
        migration_ids = self.get_migration_ids(
            migration_module_dict=migration_module_dict
        )

        query = (
            Migration.select(Migration.name)
            .where(Migration.app_name == self.app_name)
            .output(as_list=True)
        )

        if len(migration_ids) > 0:
            query = query.where(Migration.name.not_in(migration_ids))

        return t.cast(t.List[str], query.run_sync())

    async def run(self):
        print("Checking the migration table ...")

        # Make sure the migration table exists, otherwise we'll get an error.
        await self.create_migration_table()

        migration_ids_to_remove = self.get_migration_ids_to_remove()

        if migration_ids_to_remove:
            id_string = "\n".join(migration_ids_to_remove)
            print(
                "The following migrations exist in the table, not not in "
                "modules:"
            )
            print(id_string)
            confirm = (
                "y"
                if self.auto_agree
                else input("Would you like to delete these rows? (y/N)")
            )
            if confirm == "y":
                await Migration.delete().where(
                    Migration.name.is_in(migration_ids_to_remove)
                ).run()
                print("Deleted")
            else:
                print("Cancelled")
        else:
            print(
                "No migrations exist in the table, which aren't also in "
                "modules."
            )


async def clean(app_name: str, auto_agree: bool = False):
    """
    Identifies any rows in the migration table which have no corresponding
    migration module on disk, and then optionally deletes those rows.

    :param app_name:
        The name of the app to check.
    :param auto_agree:
        If true, automatically agree to any input prompts.

    """
    await CleanMigrationManager(app_name=app_name, auto_agree=auto_agree).run()



from .diffable_table import DiffableTable
from .migration_manager import MigrationManager
from .schema_differ import AlterStatements, SchemaDiffer
from .schema_snapshot import SchemaSnapshot

__all__ = [
    "DiffableTable",
    "MigrationManager",
    "AlterStatements",
    "SchemaDiffer",
    "SchemaSnapshot",
]



"""
These functions mirror their counterparts in the Postgresql docs:

https://www.postgresql.org/docs/current/functions-math.html

"""

from .base import Function


class Abs(Function):
    """
    Absolute value.
    """

    function_name = "ABS"


class Ceil(Function):
    """
    Nearest integer greater than or equal to argument.
    """

    function_name = "CEIL"


class Floor(Function):
    """
    Nearest integer less than or equal to argument.
    """

    function_name = "FLOOR"


class Round(Function):
    """
    Rounds to nearest integer.
    """

    function_name = "ROUND"


__all__ = (
    "Abs",
    "Ceil",
    "Floor",
    "Round",
)



"""
These functions mirror their counterparts in the Postgresql docs:

https://www.postgresql.org/docs/current/functions-string.html

"""

from .base import Function


class Length(Function):
    """
    Returns the number of characters in the string.
    """

    function_name = "LENGTH"


class Lower(Function):
    """
    Converts the string to all lower case, according to the rules of the
    database's locale.
    """

    function_name = "LOWER"


class Ltrim(Function):
    """
    Removes the longest string containing only characters in characters (a
    space by default) from the start of string.
    """

    function_name = "LTRIM"


class Reverse(Function):
    """
    Return reversed string.

    Not supported in SQLite.

    """

    function_name = "REVERSE"


class Rtrim(Function):
    """
    Removes the longest string containing only characters in characters (a
    space by default) from the end of string.
    """

    function_name = "RTRIM"


class Upper(Function):
    """
    Converts the string to all upper case, according to the rules of the
    database's locale.
    """

    function_name = "UPPER"


__all__ = (
    "Length",
    "Lower",
    "Ltrim",
    "Reverse",
    "Rtrim",
    "Upper",
)



from unittest import TestCase

from tests.base import engine_is
from tests.example_apps.music.tables import Manager


class TestTableStr(TestCase):
    def test_str(self):
        if engine_is("cockroach"):
            self.assertEqual(
                Manager._table_str(),
                (
                    "class Manager(Table, tablename='manager'):\n"
                    "    id = Serial(null=False, primary_key=True, unique=False, index=False, index_method=IndexMethod.btree, choices=None, db_column_name='id', secret=False)\n"  # noqa: E501
                    "    name = Varchar(length=50, default='', null=False, primary_key=False, unique=False, index=False, index_method=IndexMethod.btree, choices=None, db_column_name=None, secret=False)\n"  # noqa: E501
                ),
            )
        else:
            self.assertEqual(
                Manager._table_str(),
                (
                    "class Manager(Table, tablename='manager'):\n"
                    "    id = Serial(null=False, primary_key=True, unique=False, index=False, index_method=IndexMethod.btree, choices=None, db_column_name='id', secret=False)\n"  # noqa: E501
                    "    name = Varchar(length=50, default='', null=False, primary_key=False, unique=False, index=False, index_method=IndexMethod.btree, choices=None, db_column_name=None, secret=False)\n"  # noqa: E501
                ),
            )

        self.assertEqual(
            Manager._table_str(abbreviated=True),
            (
                "class Manager(Table):\n"
                "    id = Serial()\n"
                "    name = Varchar()\n"
            ),
        )

        # We should also be able to print it directly.
        print(Manager)



import datetime
from unittest import TestCase

from tests.base import engines_only
from tests.example_apps.music.tables import Band


@engines_only("postgres", "cockroach")
class TestArrayPostgres(TestCase):

    def test_string(self):
        self.assertEqual(
            Band.name.get_sql_value(["a", "b", "c"]),
            '\'{"a","b","c"}\'',
        )

    def test_int(self):
        self.assertEqual(
            Band.name.get_sql_value([1, 2, 3]),
            "'{1,2,3}'",
        )

    def test_nested(self):
        self.assertEqual(
            Band.name.get_sql_value([1, 2, 3, [4, 5, 6]]),
            "'{1,2,3,{4,5,6}}'",
        )

    def test_time(self):
        self.assertEqual(
            Band.name.get_sql_value([datetime.time(hour=8, minute=0)]),
            "'{\"08:00:00\"}'",
        )


@engines_only("sqlite")
class TestArraySQLite(TestCase):
    """
    Note, we use ``.replace(" ", "")`` because we serialise arrays using
    Python's json library, and there is inconsistency between Python versions
    (some output ``["a", "b", "c"]``, and others ``["a","b","c"]``).
    """

    def test_string(self):
        self.assertEqual(
            Band.name.get_sql_value(["a", "b", "c"]).replace(" ", ""),
            '\'["a","b","c"]\'',
        )

    def test_int(self):
        self.assertEqual(
            Band.name.get_sql_value([1, 2, 3]).replace(" ", ""),
            "'[1,2,3]'",
        )

    def test_nested(self):
        self.assertEqual(
            Band.name.get_sql_value([1, 2, 3, [4, 5, 6]]).replace(" ", ""),
            "'[1,2,3,[4,5,6]]'",
        )

    def test_time(self):
        self.assertEqual(
            Band.name.get_sql_value([datetime.time(hour=8, minute=0)]),
            "'[\"08:00:00\"]'",
        )



from __future__ import annotations

import typing as t

from piccolo.custom_types import Combinable, TableInstance
from piccolo.query.base import Query
from piccolo.query.methods.select import Select
from piccolo.query.mixins import WhereDelegate
from piccolo.querystring import QueryString


class Exists(Query[TableInstance, bool]):
    __slots__ = ("where_delegate",)

    def __init__(self, table: t.Type[TableInstance], **kwargs):
        super().__init__(table, **kwargs)
        self.where_delegate = WhereDelegate()

    def where(self: Self, *where: t.Union[Combinable, QueryString]) -> Self:
        self.where_delegate.where(*where)
        return self

    async def response_handler(self, response) -> bool:
        # Convert to a bool - postgres returns True, and sqlite return 1.
        return bool(response[0]["exists"])

    @property
    def default_querystrings(self) -> t.Sequence[QueryString]:
        select = Select(table=self.table)
        select.where_delegate._where = self.where_delegate._where
        return [
            QueryString(
                'SELECT EXISTS({}) AS "exists"', select.querystrings[0]
            )
        ]


Self = t.TypeVar("Self", bound=Exists)



from __future__ import annotations

import os
import sys
import typing as t

from piccolo.table import TABLE_REGISTRY


class set_env_var:
    def __init__(self, var_name: str, temp_value: str):
        """
        Temporarily set an environment variable.

        :param var_name:
            The name of the environment variable to temporarily change.
        :temp_value:
            The value that the environment variable will temporarily be set to,
            before being reset to it's pre-existing value.

        """
        self.var_name = var_name
        self.temp_value = temp_value

    def set_var(self, value: str):
        os.environ[self.var_name] = value

    def get_var(self) -> t.Optional[str]:
        return os.environ.get(self.var_name)

    def __enter__(self):
        self.existing_value = self.get_var()
        self.set_var(self.temp_value)

    def __exit__(self, *args):
        if self.existing_value is None:
            del os.environ[self.var_name]
        else:
            self.set_var(self.existing_value)


def run_pytest(pytest_args: t.List[str]) -> int:  # pragma: no cover
    try:
        import pytest
    except ImportError:
        sys.exit(
            "Couldn't find pytest. Please use `pip install 'piccolo[pytest]' "
            "to use this feature."
        )

    return pytest.main(pytest_args)


def refresh_db():
    for table_class in TABLE_REGISTRY:
        # In case any table classes were imported before we set the
        # environment variable.
        table_class._meta.refresh_db()


def run(
    pytest_args: str = "", piccolo_conf: str = "piccolo_conf_test"
) -> None:
    """
    Run your unit test suite using Pytest.

    While running, it sets the ``PICCOLO_TEST_RUNNER`` environment variable to
    ``'True'``, in case any other code needs to be aware of this.

    :param piccolo_conf:
        The piccolo_conf module to use when running your tests. This will
        contain the database settings you want to use. For example
        `my_folder.piccolo_conf_test`.
    :param pytest_args:
        Any options you want to pass to Pytest. For example
        `piccolo tester run --pytest_args="-s"`.

    """
    with set_env_var(var_name="PICCOLO_CONF", temp_value=piccolo_conf):
        refresh_db()
        args = pytest_args.split(" ")

        with set_env_var(var_name="PICCOLO_TEST_RUNNER", temp_value="True"):
            sys.exit(run_pytest(args))



from __future__ import annotations

import typing as t

from piccolo.custom_types import TableInstance
from piccolo.query.base import Query
from piccolo.querystring import QueryString


class TableExists(Query[TableInstance, bool]):

    __slots__: t.Tuple = ()

    async def response_handler(self, response):
        return bool(response[0]["exists"])

    @property
    def sqlite_querystrings(self) -> t.Sequence[QueryString]:
        return [
            QueryString(
                "SELECT EXISTS(SELECT * FROM sqlite_master WHERE "
                "name = {}) AS 'exists'",
                self.table._meta.tablename,
            )
        ]

    @property
    def postgres_querystrings(self) -> t.Sequence[QueryString]:
        subquery = QueryString(
            "SELECT * FROM information_schema.tables WHERE table_name = {}",
            self.table._meta.tablename,
        )

        if self.table._meta.schema:
            subquery = QueryString(
                "{} AND table_schema = {}", subquery, self.table._meta.schema
            )

        query = QueryString("SELECT EXISTS({})", subquery)

        return [query]

    @property
    def cockroach_querystrings(self) -> t.Sequence[QueryString]:
        return self.postgres_querystrings



from __future__ import annotations

import sys
import typing as t

from piccolo.apps.migrations.auto.migration_manager import MigrationManager
from piccolo.apps.migrations.commands.base import (
    BaseMigrationManager,
    MigrationResult,
)
from piccolo.apps.migrations.tables import Migration
from piccolo.conf.apps import AppConfig, MigrationModule
from piccolo.utils.printing import print_heading


class ForwardsMigrationManager(BaseMigrationManager):
    def __init__(
        self,
        app_name: str,
        migration_id: str = "all",
        fake: bool = False,
        preview: bool = False,
    ):
        self.app_name = app_name
        self.migration_id = migration_id
        self.fake = fake
        self.preview = preview
        super().__init__()

    async def run_migrations(self, app_config: AppConfig) -> MigrationResult:
        already_ran = await Migration.get_migrations_which_ran(
            app_name=app_config.app_name
        )

        migration_modules: t.Dict[str, MigrationModule] = (
            self.get_migration_modules(
                app_config.resolved_migrations_folder_path
            )
        )

        ids = self.get_migration_ids(migration_modules)
        n = len(ids)
        print(f"👍 {n} migration{'s' if n != 1 else ''} already complete")

        havent_run = sorted(set(ids) - set(already_ran))
        if len(havent_run) == 0:
            # Make sure this still appears successful, as we don't want this
            # to appear as an error in automated scripts.
            message = "🏁 No migrations need to be run"
            print(message)
            return MigrationResult(success=True, message=message)
        else:
            n = len(havent_run)
            print(f"⏩ {n} migration{'s' if n != 1 else ''} not yet run")

        if self.migration_id == "all":
            subset = havent_run
        elif self.migration_id == "1":
            subset = havent_run[:1]
        else:
            try:
                index = havent_run.index(self.migration_id)
            except ValueError:
                message = f"{self.migration_id} is unrecognised"
                print(message, file=sys.stderr)
                return MigrationResult(success=False, message=message)
            else:
                subset = havent_run[: index + 1]

        if subset:
            n = len(subset)
            print(f"🚀 Running {n} migration{'s' if n != 1 else ''}:")

            for _id in subset:
                if self.fake:
                    print(f"- {_id}: faked! ⏭️")
                else:
                    migration_module = migration_modules[_id]
                    response = await migration_module.forwards()

                    if isinstance(response, MigrationManager):
                        if self.preview:
                            response.preview = True
                        await response.run()

                    print("ok! ✔️")
                if not self.preview:
                    await Migration.insert().add(
                        Migration(name=_id, app_name=app_config.app_name)
                    ).run()

        return MigrationResult(success=True, message="migration succeeded")

    async def run(self) -> MigrationResult:
        await self.create_migration_table()

        app_config = self.get_app_config(app_name=self.app_name)

        return await self.run_migrations(app_config)


async def run_forwards(
    app_name: str,
    migration_id: str = "all",
    fake: bool = False,
    preview: bool = False,
) -> MigrationResult:
    """
    Run the migrations. This function can be used to programatically run
    migrations - for example, in a unit test.
    """
    if app_name == "all":
        sorted_app_names = BaseMigrationManager().get_sorted_app_names()
        for _app_name in sorted_app_names:
            print_heading(_app_name)
            manager = ForwardsMigrationManager(
                app_name=_app_name,
                migration_id="all",
                fake=fake,
                preview=preview,
            )
            response = await manager.run()
            if not response.success:
                return response

        return MigrationResult(success=True)

    else:
        manager = ForwardsMigrationManager(
            app_name=app_name,
            migration_id=migration_id,
            fake=fake,
            preview=preview,
        )
        return await manager.run()


async def forwards(
    app_name: str,
    migration_id: str = "all",
    fake: bool = False,
    preview: bool = False,
):
    """
    Runs any migrations which haven't been run yet.

    :param app_name:
        The name of the app to migrate. Specify a value of 'all' to run
        migrations for all apps.
    :param migration_id:
        Migrations will be ran up to and including this migration_id.
        Specify a value of 'all' to run all of the migrations. Specify a
        value of '1' to just run the next migration.
    :param fake:
        If set, will record the migrations as being run without actually
        running them.
    :param preview:
        If true, don't actually run the migration, just print the SQL queries

    """
    response = await run_forwards(
        app_name=app_name,
        migration_id=migration_id,
        fake=fake,
        preview=preview,
    )

    if not response.success:
        sys.exit(1)



from __future__ import annotations

import typing as t

try:
    import orjson

    ORJSON = True
except ImportError:
    import json

    ORJSON = False


def dump_json(data: t.Any, pretty: bool = False) -> str:
    if ORJSON:
        orjson_params: t.Dict[str, t.Any] = {"default": str}
        if pretty:
            orjson_params["option"] = (
                orjson.OPT_INDENT_2 | orjson.OPT_APPEND_NEWLINE  # type: ignore
            )
        return orjson.dumps(data, **orjson_params).decode(  # type: ignore
            "utf8"
        )
    else:
        params: t.Dict[str, t.Any] = {"default": str}
        if pretty:
            params["indent"] = 2
        return json.dumps(data, **params)  # type: ignore


def load_json(data: str) -> t.Any:
    return orjson.loads(data) if ORJSON else json.loads(data)  # type: ignore



import inspect
import typing as t

from typing_extensions import Protocol

from piccolo.query.base import FrozenQuery
from piccolo.utils.sync import run_sync


class Runnable(Protocol):
    async def run(
        self, node: t.Optional[str] = None, in_pool: bool = True
    ): ...


QueryType = t.TypeVar("QueryType", bound=Runnable)
ResponseType = t.TypeVar("ResponseType")


class Proxy(t.Generic[QueryType, ResponseType]):
    def __init__(self, query: QueryType):
        self.query = query

    async def run(
        self,
        node: t.Optional[str] = None,
        in_pool: bool = True,
    ) -> ResponseType:
        return await self.query.run(node=node, in_pool=in_pool)

    def run_sync(self, *args, **kwargs) -> ResponseType:
        return run_sync(self.run(*args, **kwargs))

    def __await__(
        self,
    ) -> t.Generator[None, None, ResponseType]:
        """
        If the user doesn't explicity call .run(), proxy to it as a
        convenience.
        """
        return self.run().__await__()

    def freeze(self):
        self.query.freeze()
        return FrozenQuery(query=self)

    def __getattr__(self, name: str):
        """
        Proxy any attributes to the underlying query, so all of the query
        clauses continue to work.
        """
        attr = getattr(self.query, name)

        if inspect.ismethod(attr):
            # We do this to preserve the fluent interface.

            def proxy(*args, **kwargs):
                response = attr(*args, **kwargs)
                if isinstance(response, self.query.__class__):
                    self.query = response
                    return self
                else:
                    return response

            return proxy
        else:
            return attr

    def __str__(self) -> str:
        return self.query.__str__()



from __future__ import annotations

import contextvars
import logging
import pprint
import string
import typing as t
from abc import ABCMeta, abstractmethod

from typing_extensions import Self

from piccolo.querystring import QueryString
from piccolo.utils.sync import run_sync
from piccolo.utils.warnings import Level, colored_string, colored_warning

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.query.base import DDL, Query


logger = logging.getLogger(__name__)
# This is a set to speed up lookups from O(n) when
# using str vs O(1) when using set[str]
VALID_SAVEPOINT_CHARACTERS: t.Final[set[str]] = set(
    string.ascii_letters + string.digits + "-" + "_"
)


def validate_savepoint_name(savepoint_name: str) -> None:
    """Validates a save point's name meets the required character set."""
    if not all(i in VALID_SAVEPOINT_CHARACTERS for i in savepoint_name):
        raise ValueError(
            "Savepoint names can only contain the following characters:"
            f" {VALID_SAVEPOINT_CHARACTERS}"
        )


class BaseBatch(metaclass=ABCMeta):
    @abstractmethod
    async def __aenter__(self: Self, *args, **kwargs) -> Self: ...

    @abstractmethod
    async def __aexit__(self, *args, **kwargs): ...

    @abstractmethod
    def __aiter__(self: Self) -> Self: ...

    @abstractmethod
    async def __anext__(self) -> t.List[t.Dict]: ...


class BaseTransaction(metaclass=ABCMeta):

    __slots__: t.Tuple[str, ...] = tuple()

    @abstractmethod
    async def __aenter__(self, *args, **kwargs): ...

    @abstractmethod
    async def __aexit__(self, *args, **kwargs) -> bool: ...


class BaseAtomic(metaclass=ABCMeta):

    __slots__: t.Tuple[str, ...] = tuple()

    @abstractmethod
    def add(self, *query: t.Union[Query, DDL]): ...

    @abstractmethod
    async def run(self): ...

    @abstractmethod
    def run_sync(self): ...

    @abstractmethod
    def __await__(self): ...


TransactionClass = t.TypeVar("TransactionClass", bound=BaseTransaction)


class Engine(t.Generic[TransactionClass], metaclass=ABCMeta):
    __slots__ = (
        "query_id",
        "log_queries",
        "log_responses",
        "engine_type",
        "min_version_number",
        "current_transaction",
    )

    def __init__(
        self,
        engine_type: str,
        min_version_number: t.Union[int, float],
        log_queries: bool = False,
        log_responses: bool = False,
    ):
        self.log_queries = log_queries
        self.log_responses = log_responses
        self.engine_type = engine_type
        self.min_version_number = min_version_number

        run_sync(self.check_version())
        run_sync(self.prep_database())
        self.query_id = 0

    @abstractmethod
    async def get_version(self) -> float:
        pass

    @abstractmethod
    def get_version_sync(self) -> float:
        pass

    @abstractmethod
    async def prep_database(self):
        pass

    @abstractmethod
    async def batch(
        self,
        query: Query,
        batch_size: int = 100,
        node: t.Optional[str] = None,
    ) -> BaseBatch:
        pass

    @abstractmethod
    async def run_querystring(
        self, querystring: QueryString, in_pool: bool = True
    ):
        pass

    @abstractmethod
    async def run_ddl(self, ddl: str, in_pool: bool = True):
        pass

    @abstractmethod
    def transaction(self, *args, **kwargs) -> TransactionClass:
        pass

    @abstractmethod
    def atomic(self) -> BaseAtomic:
        pass

    async def check_version(self):
        """
        Warn if the database version isn't supported.
        """
        try:
            version_number = await self.get_version()
        except Exception as exception:
            colored_warning(
                f"Unable to fetch server version: {exception}",
                level=Level.high,
            )
            return

        engine_type = self.engine_type.capitalize()
        logger.info(f"Running {engine_type} version {version_number}")
        if version_number and (version_number < self.min_version_number):
            message = (
                f"This version of {self.engine_type} isn't supported "
                f"(< {self.min_version_number}) - some features might not be "
                "available. For instructions on installing databases, see the "
                "Piccolo docs."
            )
            colored_warning(message, stacklevel=3)

    def _connection_pool_warning(self):
        message = (
            f"Connection pooling is not supported for {self.engine_type}."
        )
        logger.warning(message)
        colored_warning(message, stacklevel=3)

    async def start_connection_pool(self):
        """
        The database driver doesn't implement connection pooling.
        """
        self._connection_pool_warning()

    async def close_connection_pool(self):
        """
        The database driver doesn't implement connection pooling.
        """
        self._connection_pool_warning()

    ###########################################################################

    current_transaction: contextvars.ContextVar[t.Optional[TransactionClass]]

    def transaction_exists(self) -> bool:
        """
        Find out if a transaction is currently active.

        :returns:
            ``True`` if a transaction is already active for the current
            asyncio task. This is useful to know, because nested transactions
            aren't currently supported, so you can check if an existing
            transaction is already active, before creating a new one.

        """
        return self.current_transaction.get() is not None

    ###########################################################################
    # Logging queries and responses

    def get_query_id(self) -> int:
        self.query_id += 1
        return self.query_id

    def print_query(self, query_id: int, query: str):
        print(colored_string(f"\nQuery {query_id}:"))
        print(query)

    def print_response(self, query_id: int, response: t.List):
        print(
            colored_string(f"\nQuery {query_id} response:", level=Level.high)
        )
        pprint.pprint(response)



from piccolo.columns.combination import WhereRaw

from .base import Query
from .functions.aggregate import Avg, Max, Min, Sum
from .methods import (
    Alter,
    Count,
    Create,
    CreateIndex,
    Delete,
    DropIndex,
    Exists,
    Insert,
    Objects,
    Raw,
    Select,
    TableExists,
    Update,
)
from .methods.select import SelectRaw
from .mixins import OrderByRaw

__all__ = [
    "Alter",
    "Avg",
    "Count",
    "Create",
    "CreateIndex",
    "Delete",
    "DropIndex",
    "Exists",
    "Insert",
    "Max",
    "Min",
    "Objects",
    "OrderByRaw",
    "Query",
    "Raw",
    "Select",
    "SelectRaw",
    "Sum",
    "TableExists",
    "Update",
    "WhereRaw",
]



from piccolo.query.functions import Reverse, Upper
from piccolo.querystring import QueryString
from tests.base import engines_skip
from tests.example_apps.music.tables import Band

from .base import BandTest


@engines_skip("sqlite")
class TestNested(BandTest):
    """
    Skip the the test for SQLite, as it doesn't support ``Reverse``.
    """

    def test_nested(self):
        """
        Make sure we can nest functions.
        """
        response = Band.select(Upper(Reverse(Band.name))).run_sync()
        self.assertListEqual(response, [{"upper": "SATSINOHTYP"}])

    def test_nested_with_joined_column(self):
        """
        Make sure nested functions can be used on a column from a joined table.
        """
        response = Band.select(Upper(Reverse(Band.manager._.name))).run_sync()
        self.assertListEqual(response, [{"upper": "ODIUG"}])

    def test_nested_within_querystring(self):
        """
        If we wrap a function in a custom QueryString - make sure the columns
        are still accessible, so joins are successful.
        """
        response = Band.select(
            QueryString("CONCAT({}, '!')", Upper(Band.manager._.name)),
        ).run_sync()

        self.assertListEqual(response, [{"concat": "GUIDO!"}])


class TestWhereClause(BandTest):

    def test_where(self):
        """
        Make sure where clauses work with functions.
        """
        response = (
            Band.select(Band.name)
            .where(Upper(Band.name) == "PYTHONISTAS")
            .run_sync()
        )
        self.assertListEqual(response, [{"name": "Pythonistas"}])

    def test_where_with_joined_column(self):
        """
        Make sure where clauses work with functions, when a joined column is
        used.
        """
        response = (
            Band.select(Band.name)
            .where(Upper(Band.manager._.name) == "GUIDO")
            .run_sync()
        )
        self.assertListEqual(response, [{"name": "Pythonistas"}])



from piccolo.columns import Integer, Text, Varchar
from piccolo.query.functions import Cast, Length
from tests.example_apps.music.tables import Band, Manager

from .base import BandTest


class TestCast(BandTest):
    def test_varchar(self):
        """
        Make sure that casting to ``Varchar`` works.
        """
        response = Band.select(
            Cast(
                Band.popularity,
                as_type=Varchar(),
            )
        ).run_sync()

        self.assertListEqual(
            response,
            [{"popularity": "1000"}],
        )

    def test_text(self):
        """
        Make sure that casting to ``Text`` works.
        """
        response = Band.select(
            Cast(
                Band.popularity,
                as_type=Text(),
            )
        ).run_sync()

        self.assertListEqual(
            response,
            [{"popularity": "1000"}],
        )

    def test_integer(self):
        """
        Make sure that casting to ``Integer`` works.
        """
        Band.update({Band.name: "1111"}, force=True).run_sync()

        response = Band.select(
            Cast(
                Band.name,
                as_type=Integer(),
            )
        ).run_sync()

        self.assertListEqual(
            response,
            [{"name": 1111}],
        )

    def test_join(self):
        """
        Make sure that casting works with joins.
        """
        Manager.update({Manager.name: "1111"}, force=True).run_sync()

        response = Band.select(
            Band.name,
            Cast(
                Band.manager.name,
                as_type=Integer(),
            ),
        ).run_sync()

        self.assertListEqual(
            response,
            [
                {
                    "name": "Pythonistas",
                    "manager.name": 1111,
                }
            ],
        )

    def test_nested_inner(self):
        """
        Make sure ``Cast`` can be passed into other functions.
        """
        Band.update({Band.name: "1111"}, force=True).run_sync()

        response = Band.select(
            Length(
                Cast(
                    Band.popularity,
                    as_type=Varchar(),
                )
            )
        ).run_sync()

        self.assertListEqual(
            response,
            [{"length": 4}],
        )

    def test_nested_outer(self):
        """
        Make sure a querystring can be passed into ``Cast`` (meaning it can be
        nested).
        """
        response = Band.select(
            Cast(
                Length(Band.name),
                as_type=Varchar(),
                alias="length",
            )
        ).run_sync()

        self.assertListEqual(
            response,
            [{"length": str(len("Pythonistas"))}],
        )

    def test_where_clause(self):
        """
        Make sure ``Cast`` works in a where clause.
        """
        response = (
            Band.select(Band.name, Band.popularity)
            .where(Cast(Band.popularity, Varchar()) == "1000")
            .run_sync()
        )

        self.assertListEqual(
            response,
            [{"name": "Pythonistas", "popularity": 1000}],
        )



from unittest import TestCase

from piccolo.table import (
    create_db_tables_sync,
    create_tables,
    drop_db_tables_sync,
)
from tests.example_apps.music.tables import Band, Manager


class TestCreateDBTables(TestCase):
    def tearDown(self) -> None:
        drop_db_tables_sync(Manager, Band)

    def test_create_db_tables(self):
        """
        Make sure the tables are created in the database.
        """
        create_db_tables_sync(Manager, Band, if_not_exists=False)
        self.assertTrue(Manager.table_exists().run_sync())
        self.assertTrue(Band.table_exists().run_sync())

    def test_create_tables(self):
        """
        This is a deprecated function, which just acts as a proxy.
        """
        create_tables(Manager, Band, if_not_exists=False)
        self.assertTrue(Manager.table_exists().run_sync())
        self.assertTrue(Band.table_exists().run_sync())



from __future__ import annotations

import typing as t

from piccolo.columns import Timestamp, Varchar
from piccolo.columns.defaults.timestamp import TimestampNow
from piccolo.table import Table


class Migration(Table):
    name = Varchar(length=200)
    app_name = Varchar(length=200)
    ran_on = Timestamp(default=TimestampNow())

    @classmethod
    async def get_migrations_which_ran(
        cls, app_name: t.Optional[str] = None
    ) -> t.List[str]:
        """
        Returns the names of migrations which have already run, by inspecting
        the database.
        """
        query = cls.select(cls.name, cls.ran_on).order_by(cls.ran_on)
        if app_name is not None:
            query = query.where(cls.app_name == app_name)
        return [i["name"] for i in await query.run()]



from unittest import TestCase
from unittest.mock import MagicMock, patch

from piccolo.apps.meta.commands.version import version


class TestVersion(TestCase):
    @patch("piccolo.apps.meta.commands.version.print")
    def test_version(self, print_: MagicMock):
        version()
        print_.assert_called_once()



import typing as t

from piccolo.columns.base import Column
from piccolo.querystring import QueryString

from .base import Function


class Avg(Function):
    """
    ``AVG()`` SQL function. Column type must be numeric to run the query.

    .. code-block:: python

        await Band.select(Avg(Band.popularity))

        # We can use an alias. These two are equivalent:

        await Band.select(
            Avg(Band.popularity, alias="popularity_avg")
        )

        await Band.select(
            Avg(Band.popularity).as_alias("popularity_avg")
        )

    """

    function_name = "AVG"


class Count(QueryString):
    """
    Used in ``Select`` queries, usually in conjunction with the ``group_by``
    clause::

        >>> await Band.select(
        ...     Band.manager.name.as_alias('manager_name'),
        ...     Count(alias='band_count')
        ... ).group_by(Band.manager)
        [{'manager_name': 'Guido', 'count': 1}, ...]

    It can also be used without the ``group_by`` clause (though you may prefer
    to the :meth:`Table.count <piccolo.table.Table.count>` method instead, as
    it's more convenient)::

        >>> await Band.select(Count())
        [{'count': 3}]

    """

    def __init__(
        self,
        column: t.Optional[Column] = None,
        distinct: t.Optional[t.Sequence[Column]] = None,
        alias: str = "count",
    ):
        """
        :param column:
            If specified, the count is for non-null values in that column.
        :param distinct:
            If specified, the count is for distinct values in those columns.
        :param alias:
            The name of the value in the response::

                # These two are equivalent:

                await Band.select(
                    Band.name, Count(alias="total")
                ).group_by(Band.name)

                await Band.select(
                    Band.name,
                    Count().as_alias("total")
                ).group_by(Band.name)

        """
        if distinct and column:
            raise ValueError("Only specify `column` or `distinct`")

        if distinct:
            engine_type = distinct[0]._meta.engine_type
            if engine_type == "sqlite":
                # SQLite doesn't allow us to specify multiple columns, so
                # instead we concatenate the values.
                column_names = " || ".join("{}" for _ in distinct)
            else:
                column_names = ", ".join("{}" for _ in distinct)

            return super().__init__(
                f"COUNT(DISTINCT({column_names}))", *distinct, alias=alias
            )
        else:
            if column:
                return super().__init__("COUNT({})", column, alias=alias)
            else:
                return super().__init__("COUNT(*)", alias=alias)


class Min(Function):
    """
    ``MIN()`` SQL function.

    .. code-block:: python

        await Band.select(Min(Band.popularity))

        # We can use an alias. These two are equivalent:

        await Band.select(
            Min(Band.popularity, alias="popularity_min")
        )

        await Band.select(
            Min(Band.popularity).as_alias("popularity_min")
        )

    """

    function_name = "MIN"


class Max(Function):
    """
    ``MAX()`` SQL function.

    .. code-block:: python

        await Band.select(
            Max(Band.popularity)
        )

        # We can use an alias. These two are equivalent:

        await Band.select(
            Max(Band.popularity, alias="popularity_max")
        )

        await Band.select(
            Max(Band.popularity).as_alias("popularity_max")
        )

    """

    function_name = "MAX"


class Sum(Function):
    """
    ``SUM()`` SQL function. Column type must be numeric to run the query.

    .. code-block:: python

        await Band.select(
            Sum(Band.popularity)
        )

        # We can use an alias. These two are equivalent:

        await Band.select(
            Sum(Band.popularity, alias="popularity_sum")
        )

        await Band.select(
            Sum(Band.popularity).as_alias("popularity_sum")
        )

    """

    function_name = "SUM"


__all__ = (
    "Avg",
    "Count",
    "Min",
    "Max",
    "Sum",
)



import os
import shutil
import tempfile
from unittest import TestCase

from piccolo.apps.app.commands.new import module_exists, new


class TestModuleExists(TestCase):
    def test_module_exists(self):
        self.assertEqual(module_exists("sys"), True)
        self.assertEqual(module_exists("abc123xyz"), False)


class TestNewApp(TestCase):
    def test_new(self):
        root = tempfile.gettempdir()
        app_name = "my_app"

        app_path = os.path.join(root, app_name)

        if os.path.exists(app_path):
            shutil.rmtree(app_path)

        new(app_name=app_name, root=root)

        self.assertTrue(os.path.exists(app_path))

    def test_new_with_clashing_name(self):
        """
        Test trying to create an app with the same name as a builtin Python
        package - it shouldn't be allowed.
        """
        root = tempfile.gettempdir()
        app_name = "sys"

        with self.assertRaises(SystemExit) as context:
            new(app_name=app_name, root=root)

        exception = context.exception
        self.assertTrue(
            str(exception.code).startswith(
                "A module called sys already exists"
            )
        )



from unittest import TestCase

from piccolo.querystring import QueryString
from tests.base import postgres_only


# TODO - add more extensive tests (increased nesting and argument count).
class TestQueryString(TestCase):

    qs = QueryString(
        "SELECT id FROM band {}", QueryString("WHERE name = {}", "Pythonistas")
    )

    def test_compile_string(self):
        compiled_string, args = self.qs.compile_string()

        self.assertEqual(
            compiled_string, "SELECT id FROM band WHERE name = $1"
        )

        self.assertEqual(args, ["Pythonistas"])

    def test_string(self):
        string = self.qs.__str__()
        self.assertEqual(
            string, "SELECT id FROM band WHERE name = 'Pythonistas'"
        )

    def test_querystring_with_no_args(self):
        qs = QueryString("SELECT name FROM band")
        self.assertEqual(qs.compile_string(), ("SELECT name FROM band", []))


@postgres_only
class TestQueryStringOperators(TestCase):
    """
    Make sure basic operations can be used on ``QueryString``.
    """

    def test_add(self):
        query = QueryString("SELECT price") + 1
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price + $1", [1]),
        )

    def test_multiply(self):
        query = QueryString("SELECT price") * 2
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price * $1", [2]),
        )

    def test_divide(self):
        query = QueryString("SELECT price") / 1
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price / $1", [1]),
        )

    def test_power(self):
        query = QueryString("SELECT price") ** 2
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price ^ $1", [2]),
        )

    def test_subtract(self):
        query = QueryString("SELECT price") - 1
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price - $1", [1]),
        )

    def test_modulus(self):
        query = QueryString("SELECT price") % 1
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price % $1", [1]),
        )

    def test_like(self):
        query = QueryString("strip(name)").like("Python%")
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("strip(name) LIKE $1", ["Python%"]),
        )

    def test_ilike(self):
        query = QueryString("strip(name)").ilike("Python%")
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("strip(name) ILIKE $1", ["Python%"]),
        )

    def test_greater_than(self):
        query = QueryString("SELECT price") > 10
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price > $1", [10]),
        )

    def test_greater_equal_than(self):
        query = QueryString("SELECT price") >= 10
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price >= $1", [10]),
        )

    def test_less_than(self):
        query = QueryString("SELECT price") < 10
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price < $1", [10]),
        )

    def test_less_equal_than(self):
        query = QueryString("SELECT price") <= 10
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price <= $1", [10]),
        )

    def test_equals(self):
        query = QueryString("SELECT price") == 10
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price = $1", [10]),
        )

    def test_not_equals(self):
        query = QueryString("SELECT price") != 10
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price != $1", [10]),
        )

    def test_is_in(self):
        query = QueryString("SELECT price").is_in([10, 20, 30])
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price IN $1", [[10, 20, 30]]),
        )

    def test_not_in(self):
        query = QueryString("SELECT price").not_in([10, 20, 30])
        self.assertIsInstance(query, QueryString)
        self.assertEqual(
            query.compile_string(),
            ("SELECT price NOT IN $1", [[10, 20, 30]]),
        )
